"""
rpc.chain

Chaining together instantiation.
--------------------------------

This module provides generic support for multiple instantiation
"""
class ChainList(list):
    """
    List implementation that provides a chain method that will
    create and append to a delegated class, passing through args.
    """

    def __init__(self, klass=None):
        self._klass = klass
        list.__init__(self)

    def chain(self, *args, **kwargs):
        """
        Take the signature passed to `chain` and instantiate a new
        `self._klass` with it. Append the reesult to ourself.
        """
        self.append(self._klass(*args, **kwargs))
        return self

def client_chain(klass, *args, **kwargs):
    """
    The general form of a client chain.

    Take a class, and some starargs, instantiate the
    ChainList, and add an instance with our starargs

    Arguments:
    - `klass`: A Python class!
    """
    return ChainList(klass=klass).chain(*args, **kwargs)




"""
rpc.thrifty

"""

from thrift.protocol import TBinaryProtocol
from thrift.server import TProcessPoolServer
from thrift.transport import TSocket
from thrift.transport import TTransport

from rpc import clients, servers

class ConnectionError(Exception):
    "Failed to connect to an interface with the passed params"


def _clientmaker(service, host, port, framed=False, timeout=1):
    "Return client instance and transport for `service'"
    transport = TSocket.TSocket(host, port)
    transport.setTimeout(timeout*1000) # Milliseconds conversion
    if framed:
        transport = TTransport.TFramedTransport(transport)
    else:
        transport = TTransport.TBufferedTransport(transport)
    protocol = TBinaryProtocol.TBinaryProtocol(transport)
    client = service.Client(protocol)
    return client, transport


class Client(clients.RpcProxy):
    """
    We wrap the Thrift service client in an additional layer
    to remove the need to worry about transports & protocols etc.

    >>> client = thrifty.Client("localhost", 8888)
    """
    flavour = "Thrift"

    def __init__(self, url, service, timeout=1):
        """
        We Allow either a URI we can parse a port number from,
        or a specific port keyword argument.
        """
        self._service = service
        self.url, port = url.split(':')
        self.port = int(port)
        self.timeout = timeout
        self._client, self._transport = _clientmaker(service, self.url, self.port,
                                                     timeout=timeout)

    def __repr__(self):
        return "<{flavour} Client for {url}:{port}>".format(
            url=self.url, flavour=self.flavour, port=self.port)

    def __enter__(self):
        """
        Open the transport and return the client for Thrift contextmanagers
        """
        self._transport.open()
        return self._client

    def __exit__(self, exc, type, stack):
        """
        Close the transport
        """
        self._transport.close()
        return


class Server(servers.Server):
    """
    The Thrift server instance.

    This class wraps the creation of our Thrift server in the
    Rpc API.
    """
    flavour = "Thrift"

    def __init__(self, service, **kwargs):
        """

        Arguments:
        - `service`:
        - `**kwargs`:
        """
        self.service = service
        super(Server, self).__init__(**kwargs)
        pass

    def scaffold(self):
        """
        This function is called at the end of the base class' init.
        """
        processor = self.service.Processor(self.handler)
        sockargs = {}
        if self.host:
            sockargs['port'] = self.host
        if self.port:
            sockargs['port'] = self.port
        transport = TSocket.TServerSocket(**sockargs)
        tfactory = TTransport.TBufferedTransportFactory()
        pfactory = TBinaryProtocol.TBinaryProtocolFactory()
        self._server = TProcessPoolServer.TProcessPoolServer(processor, transport, tfactory, pfactory)
        return

    def serve(self):
        """
        Start processing incoming requests to this server
        """
        print("Serving {flavour} on {host}:{port}".format(
                flavour=self.flavour, host=self.host, port=self.port))
        self._server.serve()



# @contextlib.contextmanager
# def Client(service, host, port, framed=False):
#     """
#     Given a Thrift service, provide a client for it.
#     """
#     client, transport = _clientmaker(service, host, port, framed=framed)
#     try:
#         transport.open()
#         yield client
#     except TTransport.TTransportException:
#         raise ConnectionError(
#             "Could not connect to {interface} on {host}:{port}".format(
#                 host=host, port=port, interface=service.__name__
#                 )
#             )
#     finally:
#         transport.close()



"""
rpc.exceptions

Common exceptions across modules
"""
class IdError(Exception):
    "API endpoint returned a query with the wrong ID"

class RemoteException(Exception):
    "The remote system raised an exception of some sort"



"""
rpc.jsonp

A JSONP server for your handlers!
"""
import json

from rpc import servers

class Server(servers.HTTPServer):
    "A JSONP Server"
    flavour = "JSONP"

    def parse_response(self, request, response):
        """
        Format the response:

        Deal with the 'callback' paramater!
        """
        resp = "{callback}({data})".format(callback=request.GET['callback'],
                                           data=json.dumps(response))
        return resp





"""
Unittests for the rpc.chain module
"""
import unittest

from rpc import chains

class ChainListTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_acts_as_list(self):
        """ Basic list-like functionality should still work """
        clist = chains.ChainList()
        clist.append(0)
        with self.assertRaises(IndexError):
            clist[1] = 1
        clist += [1]
        clist.insert(2, 2)
        self.assertEqual([0, 1, 2], clist)

    def test_chain(self):
        """ Can we call the chain method """
        clist = chains.ChainList(klass=dict)
        clist.chain(boo="coo", doo="foo").chain(goo='hoo')
        self.assertEqual({'boo': 'coo', 'doo': 'foo'}, clist[0])
        self.assertEqual({'goo': 'hoo'}, clist[1])


    def tearDown(self):
        pass

class ClientChainTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_cleint_chain(self):
        """ We should take a klass and starags, returning a `ChainList`"""
        clist = chains.client_chain(dict, [('foo', 'bar')], goo='hoo')
        self.assertIsInstance(clist, chains.ChainList)
        self.assertEqual({'foo': 'bar', 'goo': 'hoo'}, clist[0])


    def tearDown(self):
        pass




if __name__ == '__main__':
    unittest.main()



"""

In addition to the normal parameters for Clients/Servers, the JSON RPC versions
contain a `verb` argument that allows you to specify either POST or GET as the
HTTP verb.
"""
import json
import uuid

import requests

from rpc import exceptions, clients, servers, chains

"""
Client Implementation
---------------------
"""

class Client(clients.RpcProxy):
    """
    This Proxy class implements a JSONRPC API.

    The timeout parameter will specify the ammount of time to wait for a call before
    raising an error.

    `verb` can be one of wither POST or GET, passed as a string and will determine
    which HTTP verb the client will use.

    >>> with Client("http://localhost:7890") as c:
    ...     print c.sayhi("Larry")
    "Hi Larry"
    """
    flavour = "JSON RPC"

    def __init__(self, url, timeout=3, verb="POST"):
        """
        Arguments:
        - `url`: string
        - `timeout`: number
        - `verb`: HTTP verb to use
        """
        self.url = url
        self.timeout = timeout
        self.verb = verb

    def __eq__(self, other):
        try:
            return self.url == other.url
        except AttributeError:
            return False

    def _get(self, headers, payload):
        """
        Make the call to a GET JSONRPC SERVER
        """
        return requests.get(self.url, params=payload, headers=headers,
                            timeout=self.timeout)

    def _post(self, headers, payload):
        """
        Make the call to a POST JSONRPC SERVER
        """
        return requests.post(self.url, data=payload, headers=headers,
                            timeout=self.timeout)

    def _build_payload(self, *args, **kwargs):
        """
        Build the Payload for our call.

        Largely factored out as a convenient Hook fucntions
        """
        reqid = uuid.uuid4().hex
        method = args[1]
        params = args[2:]
        if kwargs:
            raise ValueError("Keyword arguments not supported by JSON RPC try passing a dict.")
        payload = dict(params=params, id=reqid, method=method)
        return reqid, {k: json.dumps(v) for k, v in payload.items()}

    def _apicall(self, *args, **kwargs):
        """
        Make a JSONRPC call to a JSONRPC server

        Arguments:
        - `data`: string
        """
        reqid, payload = self._build_payload(*args, **kwargs)
        headers = {'X-flavour': 'JSONRPC'}
        if self.verb == "GET":
            resp = self._get(headers, payload)
        elif self.verb == "POST":
            resp = self._post(headers, payload)
        else:
            raise ValueError("Unsupported HTTP Verb {verb}".format(verb=self.verb))
        result = json.loads(resp.text)
        if reqid != result['id']:
            raise exceptions.IdError("API Endpoint returned with id:{ret}, expecting:{exp}".format(
                ret=result['id'], exp=reqid))
        del result['id']
        if resp.status_code == 200:
            return result
        else:
            raise exceptions.RemoteException(result['result'])


def chain(*args, **kwargs ):
    """
    Will return an iterable which can be .chain()'ed as much as you
    like to create multiple Clients.

    >>> chain("localhost").chain("example.com")
    ... [<JSON RPC Client for localhost>, <JSON RPC Client for example.com>]
    """
    return chains.client_chain(Client, *args, **kwargs)

"""
Server Implementation
---------------------
"""

class Server(servers.HTTPServer):
    """
    A JSONRPC server

    >>> class Handler(object):
    ...     def sayhi(self, person):
    ...         return "Hi {0}".format(person)
    ...
    >>> server = Server("localhost", 7890, Handler)
    >>> server.serve()

    """
    flavour = "JSON RPC"

    def procedure(self, request):
        """
        JSON RPC procedure call - parse the params, call the procedure, and
        return the appropriate values.

        the procedure() method of HTTP Servers should return
        status, headers, content
        """
        status = '200 OK'
        headers = [('Content-Type', 'application/json')]
        result, error = None, None
        data = getattr(request, request.method)
        method, params, reqid = [json.loads(v) for v in [data.get('method', 'null'),
                                                         data.get('params', '[]'),
                                                         data.get('id', 'null')]]
        if not method:
            error = "No Method specified"
            return status, headers, dict(id=reqid, result=None, error=error)
        if not hasattr(self.handler, method):
            error = 'Method "{0}"" Not Found... '.format(method)
        if error:
            return status, headers, dict(id=id, result=result, error=error)
        try:
            result = getattr(self.handler, method)(*params)
        except Exception as err:
            error = '{error}: {msg}'.format(
                error=err.__class__.__name__, msg=err.message)
        return status, headers, dict(id=reqid, result=result, error=error)

    def parse_response(self, request, response):
        """
        Format the response:

        Just json.dump it
        """
        return json.dumps(response)



"""
Unittests for the rpc.clients module
"""
import unittest

from rpc import clients


class ClientTestCase(unittest.TestCase):

    def test_repr(self):
        """ Stringify nicely """
        client = clients.RpcProxy()
        self.assertEqual("<Base Client for None>", str(client))

    def test_contextmanager(self):
        """ Use client as a contextmanager """
        with clients.RpcProxy() as client:
            self.assertIsInstance(client, clients.RpcProxy)

    def test_getattr(self):
        """ Test the call substitution """
        client = clients.RpcProxy()
        self.assertEqual("None", client.url)
        self.assertEqual("Base", client.flavour)
        with self.assertRaises(NotImplementedError):
            client.callit(True)

    def test_apicall_nimp(self):
        """ Base apicall should raise """
        client = clients.RpcProxy()
        with self.assertRaises(NotImplementedError):
            client._apicall(False)


if __name__ == '__main__':
    unittest.main()



"""
rpc.clients

Base classes for all local client proxies to inherit from.
"""

class RpcProxy(object):
    """
    A base implementation of the proxy pattern for RPC clients.
    """
    def __init__(self):
        self.url = "None"
        self.flavour = "Base"

    def __repr__(self):
        return "<{flavour} Client for {url}>".format(
            url=self.url, flavour=self.flavour)

    def __enter__(self):
        return self

    def __exit__(self, exc, type, stack):
        return

    def __getattr__(self, key):
        """
        We allow anything not in self.__dict__ to be called as a method.
        abstracting the reqests away.
        """
        if key in self.__dict__:
            return self.__dict__[key]
        return lambda *a, **kw: self._apicall(self, key, *a, **kw)

    def _apicall(self, *args, **kwargs):
        raise NotImplementedError("No proxy dispatch method!")




"""
rpc.servers

Base class for server implementations
"""
import functools
from wsgiref import simple_server

import webob

def webobify(fn):
    "Decorator to convert a WSGI environ into a WebOb Request"

    @functools.wraps(fn)
    def munger(self, environ, start):
        return fn(self, webob.Request(environ), start)

    return munger

class Server(object):
    """
    Base class for servers.
    """

    def __init__(self, host=None, port=None, handler=None):
        """
        Arguments:
        - `host`: string
        - `port`: int
        - `handler`: callable
        """
        self.host = host
        self.port = port
        self.handler = handler()
        self.scaffold()

    def __repr__(self):
        return "<{flavour} Server on {host}:{port} calling {handler}>".format(
            flavour=self.flavour, host=self.host, port=self.port,
            handler=self.handler.__class__.__name__)

    def scaffold(self):
        """
        This hook function is called with no arguments as the last item
        of initialisation. This is an excellent subclass choice for instance
        building code.

        In this base class, is simply a no-op
        """
        return

    def serve(self):
        """
        Subclasses should override this base method to accept incoming
        calls and deal with marshalling/dispatch.
        """
        raise NotImplementedError()


class HTTPServer(Server):
    "WSGI HTTP Server"

    def __init__(self, *args, **kwargs):
        super(HTTPServer, self).__init__(*args, **kwargs)
        self.httpd = simple_server.make_server(self.host, self.port, self.app)

    def parse_response(self, request, response):
        """
        No-op hook for subclasses to override.
        """
        return response

    @webobify
    def app(self, request, start_response):
        """
        Our JSON RPC WSGI App.

        Decode and deserialize the POST data, locate the handler method,
        ascertain the result and then return our JSON response.
        """
        if request.method not in ['GET', 'POST']:
            return ["Invalid HTTP Verb {verb}".format(verb=request.method)]
        status, headers, response = self.procedure(request)
        start_response(status, headers)
        return [self.parse_response(request, response)]

    def serve(self):
        """
        Start handling requests.

        It a Sub-Optimal idea to use this in any kind of production setting.
        """
        print("Serving {flavour} on {host}:{port}".format(
                flavour=self.flavour, host=self.host, port=self.port))
        self.httpd.serve_forever()




"""
Unittests for the rpc.servers module
"""
import unittest

from rpc import servers

class ServerTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_initialise(self):
        """ Can we initialise?"""
        server = servers.Server(host="localhost", port=6786, handler=dict)
        self.assertEqual("localhost", server.host)
        self.assertEqual(6786, server.port)
        self.assertIsInstance(server.handler, dict)

    def test_scaffold(self):
        """
        Should be a no-op
        """
        server = servers.Server(host="localhost", port=6786, handler=dict)
        self.assertEqual(None, server.scaffold())

    def test_serve_raises(self):
        """ Dummy serve() should raise """
        server = servers.Server(host="localhost", port=6786, handler=dict)
        with self.assertRaises(NotImplementedError):
            server.serve()

    def tearDown(self):
        pass



if __name__ == '__main__':
    unittest.main()



import re

from distutils.core import setup

VERSION_FILE = "rpc/_version.py"
verstrline = open(VERSION_FILE, "rt").read()
VSRE = r'^__version__ = [\'"]([^\'"]*)[\'"]'
mo = re.search(VSRE,  verstrline, re.M)
if mo:
    VERSION = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in {0}".format(VERSION_FILE))

LONG_DESC = """
Rpc gives you a consistent API for RPC protocols that doesn't suck!

Because you too can have nice things::

    from rpc import thrifty

    import Service

    with thrifty.client("localhost:45678", Service) as c:
        print c.ping()
"""

setup(
    name = "rpc",
    version = VERSION,
    author = "David Miller",
    author_email = "david@deadpansincerity.com",
    url = "http://www.deadpansincerity.com/docs/rpc",
    description = "RPC Client/Server library",
    long_description = LONG_DESC,
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries"
        ],
    packages = ['rpc'],
    install_requires = ["WebOb==1.2b3",
                "requests==0.10.6"]
    )



"""
Unittests for the rpc.thrifty module
"""
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../examples/gen-py')))

import unittest

from mock import patch, Mock

from thrift.server import TProcessPoolServer
from thrift.transport import TSocket, TTransport

from service import Service
from rpc import thrifty

class ClientMakerTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_clientmaker(self):
        """ """
        c, t = thrifty._clientmaker(Service, "localhost", 567)
        self.assertIsInstance(c, Service.Client)
        self.assertIsInstance(t, TTransport.TBufferedTransport)

    def test_timeout(self):
        """ Test that the Timeout is set """
        with patch.object(TSocket.TSocket, "setTimeout") as Pset:
            c, t = thrifty._clientmaker(Service, "localhost", 666)
            Pset.assert_called_with(1000)
            c, t = thrifty._clientmaker(Service, "localhost", 666, timeout=5)
            Pset.assert_called_with(5000)

    def tearDown(self):
        pass


class ClientTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_init(self):
        """ Chain JSON RPC Clients"""
        client = thrifty.Client("localhost:30303", Service, timeout=0.5)
        self.assertEqual(Service, client._service)
        self.assertEqual("localhost", client.url)
        self.assertEqual(0.5, client.timeout)
        self.assertEqual(30303, client.port)
        self.assertIsInstance(client._transport, TTransport.TBufferedTransport)
        self.assertIsInstance(client._client, Service.Client)

    def test_repr(self):
        """ String Repr """
        client = thrifty.Client("localhost:30303", Service, timeout=0.5)
        self.assertEqual("<Thrift Client for localhost:30303>", str(client))

    def test_contextmanager(self):
        """ Use as a contextmanager """
        with patch.object(thrifty, 'TTransport') as Ptrans:
            with thrifty.Client("localhost:30303", Service) as c:
                self.assertIsInstance(c, Service.Client)
                Ptrans.TBufferedTransport.return_value.open.assert_called_once_with()
            Ptrans.TBufferedTransport.return_value.close.assert_called_once_with()

    def tearDown(self):
        pass

class ServerTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_repr(self):
        """ """
        server = thrifty.Server(host="localhost", port=4444, handler=dict, service=Service)
        exp = "<Thrift Server on localhost:4444 calling dict>"
        self.assertEqual(exp, str(server))

    def test_scaffold(self):
        """ Scaffold the server """
        server = thrifty.Server(host="localhost", port=4444, handler=dict, service=Service)
        self.assertIsInstance(server.handler, dict)
        self.assertIsInstance(server._server, TProcessPoolServer.TProcessPoolServer)

    def test_serve(self):
        "Serve should delegate"
        server = thrifty.Server(host="localhost", port=4444, handler=dict, service=Service)
        mock_serv = Mock(name='mock Tserver')
        server._server = mock_serv
        server.serve()
        mock_serv.serve.assert_called_once_with()

    def tearDown(self):
        pass




if __name__ == '__main__':
    unittest.main()



"""
Unittests for the rpc.jsonrpc module
"""
import unittest

from rpc import jsonrpc

class ChainTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_chain(self):
        """ Chain JSON RPC Clients"""
        one, two = jsonrpc.chain("localhost").chain("example.com")
        self.assertEqual(one.url, "localhost")
        self.assertEqual(two.url, "example.com")

    def tearDown(self):
        pass



if __name__ == '__main__':
    unittest.main()
